import re
from datetime import datetime, timezone
from pathlib import Path
from typing import List, Optional, Dict
from uuid import uuid4

import click
import sqlite_utils
from pydantic import BaseModel, root_validator, PrivateAttr, Extra, validator
from pydantic import Field
from rich import filesize


class Archive(BaseModel):
    sessionId: Optional[str]
    archiveId: str
    principalId: str
    username: str
    deviceId: str
    deviceName: str
    destinationId: str
    destinationName: str
    storageNodeWebUrl: str

    @property
    def json_filename(self):
        return (
            f"{self.username}_{self.deviceName}_{self.deviceId}_{self.archiveId}.json"
        )

    @property
    def db_filename(self):
        return f"{self.username}_{self.deviceName}_{self.deviceId}.db"


class FileMetadata(BaseModel, extra=Extra.ignore):
    archiveId: str
    fileId: str
    backupSetId: int
    fileType: str
    path: str
    versionTimeMillis: int
    sourceLastModifiedMillis: int
    sourceLengthBytes: int
    sourceSHA256Checksum: str
    deleted: bool
    deletionTimeMillis: Optional[int]
    selected: bool = False
    completed: bool = False
    excludeReason: Optional[str]

    @root_validator(pre=True)
    def handle_deleted_file(cls, values: dict):  # noqa
        if values["deleted"]:
            last_non_deleted = values.pop("lastNonDeletedEntry")
            values["deletionTimeMillis"] = values["versionTimeMillis"]
            values.update(last_non_deleted)
        return values


class RestorableFile(BaseModel):
    deviceId: str
    archiveId: str
    storageNodeWebUrl: str
    destinationName: str
    fileId: str
    path: str
    versionTimeMillis: int
    sourceLastModifiedMillis: int
    sourceLengthBytes: int
    sourceSHA256Checksum: str
    backupSetId: int
    error: Optional[str]
    failure_count: int = 0
    completed: bool = False

    @property
    def restore_path(self):
        # Convert Windows drive to letter folder
        value = re.sub(r"^([A-Z]):", r"\1", self.path)
        # remove root slash "/" if present
        value = value.lstrip("/")
        return f"{self.deviceId}/{value}"

    @property
    def size(self):
        """Human-readable file size. 1KB == 1000 Bytes."""
        return filesize.decimal(self.sourceLengthBytes)


class FileSelection(BaseModel):
    includeDeleted: bool = False
    backupSetIds: Optional[List[int]]
    deletedMaxAge: Optional[datetime]
    restoreAsOf: Optional[datetime]
    includePatterns: List[str] = Field(default_factory=list)
    excludePatterns: List[str] = Field(default_factory=list)

    @validator("includePatterns")
    def validate_include_regexes(cls, value):
        try:
            [re.compile(v) for v in value]
            return value
        except re.error as regex_error:
            raise click.ClickException(
                f"Unable to parse regex for include pattern '{regex_error.pattern}': {regex_error}"
            )

    @validator("excludePatterns")
    def validate_exclude_regexes(cls, value):
        try:
            [re.compile(v) for v in value]
            return value
        except re.error as regex_error:
            raise click.ClickException(
                f"Unable to parse regex for exclude pattern '{regex_error.pattern}': {regex_error}"
            )

    @property
    def endTimeMillis(self):
        if self.restoreAsOf is not None:
            return int(self.restoreAsOf.astimezone(timezone.utc).timestamp() * 1000)

    @property
    def deletedAgeMillis(self):
        if self.deletedMaxAge is not None:
            return int(self.deletedMaxAge.astimezone(timezone.utc).timestamp() * 1000)


class Device:
    def __init__(self, username: str, name: str, deviceId: str, db_file: Path):
        self.username = username
        self.name = name
        self.deviceId = deviceId
        self.db = sqlite_utils.Database(db_file)
        self.progress = None
        self.archive_progress = None
        self.task_id = None

    @property
    def completedFiles(self):
        result = next(
            self.db.query(
                """
                SELECT coalesce(count(fileId), 0) as completedFiles
                FROM file
                WHERE fileType != 'DIRECTORY'
                AND selected
                AND completed
                """
            )
        )
        return result["completedFiles"]

    @property
    def completedBytes(self):
        result = next(
            self.db.query(
                """
                SELECT coalesce(sum(sourceLengthBytes), 0) as completedBytes
                FROM file
                WHERE fileType != 'DIRECTORY'
                AND selected
                AND completed
                """
            )
        )
        return result["completedBytes"]

    @property
    def totalFiles(self):
        result = next(
            self.db.query(
                """
                SELECT coalesce(count(distinct(fileId)), 0) as totalFiles
                FROM file
                WHERE fileType != 'DIRECTORY' AND selected
                """
            )
        )
        return result["totalFiles"]

    @property
    def totalBytes(self):
        result = next(
            self.db.query(
                """
                with deduplicated as (
                    SELECT sourceLengthBytes, max(sourceLastModifiedMillis) as sourceLastModifiedMillis
                    FROM file
                    WHERE fileType != 'DIRECTORY' AND selected
                    GROUP BY path
                )
                SELECT coalesce(sum(sourceLengthBytes), 0) as totalBytes
                FROM deduplicated
                """
            )
        )
        return result["totalBytes"]


class RestoreJob(BaseModel):
    jobId: str = Field(default_factory=lambda: str(uuid4()))
    type: str
    identifier: str
    selection: FileSelection = Field(default_factory=FileSelection)
    targetFolder: Path = Field(default_factory=lambda: Path().absolute())
    archives: List[Archive]
    totalFileCount: int = 0
    totalByteCount: int = 0
    _devices: Dict[str, Device] = PrivateAttr(default_factory=dict)

    @property
    def devices(self):
        if not self._devices:
            for archive in self.archives:
                if archive.deviceId not in self._devices:
                    self._devices[archive.deviceId] = Device(
                        username=archive.username,
                        name=archive.deviceName,
                        deviceId=archive.deviceId,
                        db_file=self.folder / archive.db_filename,
                    )
        return self._devices

    @property
    def filename(self):
        return f"restore_{self.jobId}.json"

    @property
    def folder(self):
        return self.targetFolder / self.jobId

    def calculate_stats(self):
        self.totalFileCount = self.totalByteCount = 0
        for device in self.devices.values():
            self.totalFileCount += device.totalFiles
            self.totalByteCount += device.totalBytes



from typing import Generator

import httpx


class UserAuth(httpx.Auth):
    requires_response_body = True

    def __init__(self, url: str, username: str, password: str):
        self.url = url
        self.basic_auth = httpx.BasicAuth(username, password)
        self.access_token = None

    def auth_flow(self, request) -> Generator[httpx.Request, httpx.Response, None]:
        if self.access_token is None:
            response = yield self.get_auth_token()
            response.raise_for_status()
            self.access_token = response.json()["data"]["v3_user_token"]
        request.headers["Authorization"] = f"Bearer {self.access_token}"
        yield request

    def get_auth_token(self) -> httpx.Request:
        request = httpx.Request(
            url=f"{self.url}/api/v3/auth/jwt?useBody=true",
            method="GET",
        )
        return next(self.basic_auth.auth_flow(request))


class APIClientAuth(httpx.Auth):
    requires_response_body = True

    def __init__(self, url: str, api_client_id: str, api_client_secret: str):
        self.url = url
        self.basic_auth = httpx.BasicAuth(api_client_id, api_client_secret)
        self.access_token = None

    def auth_flow(self, request) -> Generator[httpx.Request, httpx.Response, None]:
        if self.access_token is None:
            response = yield self.get_auth_token()
            response.raise_for_status()
            self.access_token = response.json()["access_token"]
        request.headers["Authorization"] = f"Bearer {self.access_token}"
        yield request

    def get_auth_token(self) -> httpx.Request:
        request = httpx.Request(
            url=f"{self.url}/api/v3/oauth/token?grant_type=client_credentials",
            method="POST",
        )
        return next(self.basic_auth.auth_flow(request))



import asyncio
import json
from hashlib import sha256
from os import utime
from pathlib import Path
from queue import Queue

import click
import httpx
from httpx import HTTPStatusError
from rich.console import Console
from rich import filesize

from bulk_restore_tool.__about__ import __version__
from bulk_restore_tool.client import Client, DetailType
from bulk_restore_tool.models import (
    RestoreJob,
    FileSelection,
    RestorableFile,
    Archive,
)
from bulk_restore_tool.db import (
    save_archive_metadata_sqlite,
    restore_queue,
)
from bulk_restore_tool.progress import ProgressBar

console = Console()


def store_in_context(ctx, param, arg):
    """
    Callback for `click.option`s that stores option value in the `ctx.obj` (also ensures this is a dict).

    Useful for enabling "global" options that can be passed at any stage of the command or sub-commands.
    """
    if arg:
        ctx.ensure_object(dict)
        ctx.obj.update({param.name: arg})
    return arg


log_file_option = click.option(
    "--log-file",
    help="Specify location where logging gets written. Can also be set by BRT_LOG_FILE environment variable. "
    "Default='bulk_restore.log' in current directory.",
    callback=store_in_context,
)
log_level_option = click.option(
    "--log-level",
    type=click.Choice(["WARN", "INFO", "DEBUG"]),
    help="Specify logging level. Can also be set by BRT_LOG_LEVEL environment variable. Defaults to INFO.",
    callback=store_in_context,
)


def logging_options(f):
    f = log_file_option(f)
    f = log_level_option(f)
    return f


@click.group(
    context_settings={
        "help_option_names": ["-h", "--help"],
        "max_content_width": 200,
    },
    invoke_without_command=True,
)
@click.option(
    "--version", is_flag=True, help="Print the currently installed version of the tool."
)
@logging_options
def brt(version, **kwargs):
    """
    \b
    Code42 Bulk Restore Tool

    To authenticate, set the following variables either in your shell environment or an .env file in your
    current working directory:

    \b
        BRT_URL
        BRT_API_CLIENT_ID
        BRT_API_CLIENT_SECRET

    For details on API Client authentication, see: https://support.code42.com/Incydr/Admin/Code42_console_reference/API_clients
    """
    if version:
        console.print(__version__)
        return


@brt.command()
@click.argument("id")
@click.option(
    "--name",
    help="Specify a name for the restore job. This name will be visible in audit log entries for the restore session. "
    "Default = A generated UUID.",
)
@click.option(
    "--type",
    "type_",
    type=click.Choice([d.value for d in DetailType]),
    default="archiveId",
    help="Specify what type of identifier to gather archive metadata for. Default = archiveId",
)
@click.option(
    "--target",
    type=click.Path(),
    help="Specify target folder where archive metadata will be stored (will be created if it doesn't exist). "
    "Default = Directory `brt` is executed in.",
)
@click.option(
    "--include-pattern",
    multiple=True,
    type=str,
    help="Specify regex pattern to select which file paths to include in the restore. Can be passed multiple times.",
)
@click.option(
    "--exclude-pattern",
    multiple=True,
    type=str,
    help="Specify regex pattern to select which file paths to exclude in the restore. Can be passed multiple times. A "
    "match on an exclude pattern will take precedence over any matching include patterns.",
)
@click.option(
    "--include-deleted",
    is_flag=True,
    help="Include files that have been marked as deleted in the archive file metadata. Default = false",
)
@click.option(
    "--max-deleted-date",
    type=click.DateTime(),
    help="Specify date for maximum age of deleted files to restore. Any files that were marked as deleted before this "
    "date will be excluded from restore. Default = None",
)
@click.option(
    "--as-of-date",
    type=click.DateTime(),
    help="Restore files as of specified date. Default = None",
)
@click.option(
    "--backup-set-ids",
    type=int,
    multiple=True,
    help="Specify which backup sets from the archive should be included, can be passed multiple times to include "
    "files from multiple sets. Default = Include files from all backup sets.",
)
@click.option(
    "--no-calculate",
    is_flag=True,
    help="Skip gathering of archive metadata and calculation of restore totals and only create job file. "
    "Default = false",
)
@logging_options
def create_job(
    id,
    target,
    type_,
    include_pattern,
    exclude_pattern,
    include_deleted,
    backup_set_ids,
    max_deleted_date,
    as_of_date,
    log_level,
    log_file,
    name,
    no_calculate,
):
    """
    Creates a restore job for a set of archives.

    Archives can be included in the job by specifying the identifier type (archiveId, deviceId, username, or legalHoldId)
    via the --type option (defaults to archiveId) and providing a comma-separated list of identifiers as the main command-line
    argument. Some examples:

        \b
        brt create-job --type archiveId id1,id2,id3
        brt create-job --type username user@example.com
        brt create-job --type deviceId 1234,4567
        brt create-job --type legalHoldId 7890

    If the --name option is not provided, a new UUID will be created as the job name and the parameters for the job will
    be saved in restore_<job_name>.json

    A <job_name> folder will also be created in the --target path to store the archive metadata and restored files (if
    --target option is not provided, it defaults to the current working directory).

    In the <job_name> folder, a human-readable .json file will be generated for each archive containing the file
    metadata from that archive. Additionally, a sqlite .db file will also be generated for each device in the job to
    track restore progress per device. Because a device can have multiple (destination) archives and/or multiple backup
    sets within a single archive, it's possible that some file paths will be duplicated within the full set of a single
    device's metadata. While duplicates will remain in the individual archive metadata files, they will be de-duplicated
    when restoring and when calculating the total file count and byte count of the restore job (the fileId with the most
    recent modification timestamp will be restored).
    """
    target = Path(target) if target else Path(".")
    params = {
        "includeDeleted": include_deleted,
        "backupSetIds": backup_set_ids,
        "deletedMaxAge": max_deleted_date,
        "restoreAsOf": as_of_date,
        "includePatterns": include_pattern,
        "excludePatterns": exclude_pattern,
    }
    client = Client(log_level=log_level, log_file=log_file)
    ids = id.split(",")
    archives = []
    for id in ids:
        archives.extend(client.get_archive_details(id, id_type=type_))
    job = RestoreJob(
        type=type_,
        identifier=id,
        selection=FileSelection(**params),
        archives=archives,
        targetFolder=target.absolute(),
    )
    if name:
        job.jobId = name
    if no_calculate:
        with open(f"restore_{job.jobId}.json", "w") as f:
            f.write(job.json(indent=2))
    else:
        _prepare_job(job, client)


def _prepare_job(job: RestoreJob, client: Client):
    console.print("Creating job folders...")
    job.folder.mkdir(parents=True, exist_ok=True)

    selection_params = {
        "endTime": job.selection.endTimeMillis,
        "includeDeleted": job.selection.includeDeleted,
    }
    msg = f"Fetching archive metadata with the following selection criteria: {job.selection}"
    console.print(msg)
    client.settings.logger.info(msg)
    for archive in job.archives:
        console.print(
            f"  user={archive.username}, device_name={archive.deviceName}, device_guid={archive.deviceId}, archive={archive.archiveId}"
        )
        manifest = client.get_archive_metadata(archive, **selection_params).json()
        with open(job.folder / archive.json_filename, "w") as f:
            json.dump(manifest, f, indent=2)
        files = manifest["data"]["archiveMetadataEntries"]
        save_archive_metadata_sqlite(archive, files, job.folder, job.selection)

    console.print("Calculating restore size...")
    job.calculate_stats()
    msg = f"Job size is {job.totalFileCount} files totaling {filesize.decimal(job.totalByteCount)}"
    console.print(msg)
    client.settings.logger.info(msg)
    with open(job.filename, "w") as f:
        f.write(job.json(indent=2))
        msg = f"Job definition saved: {job.filename}"
        console.print(msg)
        client.settings.logger.info(msg)


@brt.command()
@click.argument("JOB_FILE", type=click.File())
def start_job(job_file):
    """Begins restoring from archives defined in a given job file."""
    client = Client()
    job = RestoreJob.parse_raw(job_file.read())
    if not job.folder.exists():
        _prepare_job(job, client)
    else:
        refresh = False
        for archive in job.archives:
            if not (job.folder / archive.json_filename).exists():
                msg = f"Metadata files missing for archiveId {archive.archiveId}..."
                console.print(msg)
                client.settings.logger.warning(msg)
                refresh = True
        if refresh:
            msg = "Re-fetching job metadata..."
            console.print(msg)
            client.settings.logger.warning(msg)
            _prepare_job(job, client)
    for archive in job.archives:
        if archive.sessionId is None:
            archive.sessionId = client.create_bulk_restore_session(
                archive, name=job.jobId
            )
            client.settings.logger.info(
                f"Created bulk restore sessionId={archive.sessionId} for archiveId={archive.archiveId} on server={archive.storageNodeWebUrl}"
            )
    with open(job.filename, "w") as f:
        f.write(job.json(indent=2))
    asyncio.run(restore_main(job, client))


SESSIONS_PER_ARCHIVE = 1


async def restore_main(job: RestoreJob, client: Client):
    with ProgressBar(job) as progress_bar:
        tasks = []
        for archive in job.archives:
            db = job.devices[archive.deviceId].db
            file_queue = restore_queue(db, archive.archiveId)
            for _ in range(SESSIONS_PER_ARCHIVE):
                task = restore_archive(
                    client=client,
                    archive=archive,
                    job=job,
                    file_queue=file_queue,
                    progress_bar=progress_bar,
                )
                tasks.append(task)
        for completed in asyncio.as_completed(tasks):
            archive = await completed
            try:
                client.complete_bulk_restore_session(archive, archive.sessionId)  # type: ignore
            except HTTPStatusError:
                ...
            msg = f"Completed all queued files for archive {archive.archiveId} of device {archive.deviceId}."
            client.settings.logger.info(msg)


async def restore_archive(
    client: Client,
    archive: Archive,
    job: RestoreJob,
    file_queue: Queue,
    progress_bar: ProgressBar,
):
    device = job.devices[archive.deviceId]

    # A device/error logger for each device writes to its own file for easy audit of the restore record for each device.
    # The error logger will only log files that failed all 3 restore attempts so an end-user can quickly identify
    # which files are not in the target folder that _should_ be.
    # The device loggers also write to the main log file (if the --log-level option is INFO or higher)
    device_logger, error_logger = client.settings.get_device_loggers(device, job.folder)
    main_logger = client.settings.logger

    file_table = device.db["file"]
    storage_client = httpx.AsyncClient(
        auth=client.config["auth"], base_url=archive.storageNodeWebUrl, timeout=600  # type: ignore
    )
    while not file_queue.empty():
        file: RestorableFile = file_queue.get()
        progress_bar.start_file_progress(file)
        try:
            await restore_file(storage_client, archive, file, job.folder)
            file_table.update(
                pk_values=(file.archiveId, file.fileId), updates={"completed": True}
            )
            file.completed = True
            device_logger.info(
                f"Restored archive={archive.archiveId} fileId={file.fileId} sourcePath={file.path} targetPath={job.folder / file.restore_path} bytes={file.sourceLengthBytes} ({file.size})"
            )

        except InvalidBulkRestoreSession:
            main_logger.warning(
                f"Invalid bulk restore sessionId={archive.sessionId}, creating new session."
            )
            archive.sessionId = client.create_bulk_restore_session(
                archive, name=job.jobId
            )
            main_logger.debug(f"Updating job file: {job.json()}")
            with open(job.filename, "w") as f:
                f.write(job.json(indent=2))
            file_queue.queue.appendleft(file)
            progress_bar.complete_file_progress(file)
            continue

        except Exception as err:
            # track failed attempts, log each error and retry up to `max_attempts` times
            max_attempts = 3
            file.failure_count += 1
            device_logger.warning(
                f"Failed attempt {file.failure_count}/{max_attempts} to restore fileId={file.fileId} path={file.path}. Error: {err}"
            )
            client.settings.logger.debug(err, exc_info=True)

            if file.failure_count < max_attempts:
                file_queue.queue.appendleft(file)
            else:
                error_logger.warning(
                    f"{max_attempts} restore attempts failed for {file.path} (fileId={file.fileId}) from archiveId={archive.archiveId} on server={archive.storageNodeWebUrl}"
                )

        progress_bar.complete_file_progress(file)
    return archive


async def restore_file(
    client: httpx.AsyncClient, archive: Archive, file: RestorableFile, target: Path
):
    file_path = target / file.restore_path
    file_path.parent.mkdir(parents=True, exist_ok=True)
    if file.sourceLengthBytes == 0:
        # don't make network requests for 0byte files
        file_path.touch(exist_ok=True)
        return
    params = {
        "bulkRestoreSessionId": archive.sessionId,
        "archiveGuid": archive.archiveId,
        "fileId": file.fileId,
        "versionTimestamp": file.versionTimeMillis,
    }
    with open(file_path, "wb") as target_file:
        async with client.stream(
            method="GET", url="/api/v3/bulk-restore-file-data", params=params
        ) as response:
            if (
                response.status_code == 404
                and "Bulk Restore Session ID not found" in response.text
            ):
                raise InvalidBulkRestoreSession()
            response.raise_for_status()
            async for chunk in response.aiter_bytes(chunk_size=1024):
                target_file.write(chunk)
    # apply last modified/access metadata
    sourceLastModifiedNanoSeconds = file.sourceLastModifiedMillis * 1_000_000
    utime(file_path, ns=(sourceLastModifiedNanoSeconds, sourceLastModifiedNanoSeconds))

    checksum = sha256(file_path.read_bytes()).hexdigest()
    if checksum != file.sourceSHA256Checksum:
        raise Exception(
            f"Invalid SHA256 checksum for fileId={file.fileId}, expected {file.sourceSHA256Checksum}, got {checksum}."
        )


class InvalidBulkRestoreSession(Exception):
    ...



import logging
import time
from pathlib import Path
from typing import Optional

from pydantic import BaseSettings
from pydantic import SecretStr
from pydantic import validator

from click.types import LazyFile

from bulk_restore_tool.models import Device


class UTCFormatter(logging.Formatter):
    converter = time.gmtime


brt_format = UTCFormatter(
    fmt="%(asctime)s %(levelname)s %(name)s - %(message)s", datefmt="%Y-%m-%dT%H:%M:%SZ"
)


class BRTSettings(BaseSettings):
    """
    Setting configuration class for the Bulk Restore Tool.

    Inheriting from [pydantic.BaseSettings](https://pydantic-docs.helpmanual.io/usage/settings/) enables environment
    variable and .env file fallbacks if values are not passed in during instantiation, allowing flexible configuration
    from multiple possible sources, which-ever is easier for the end-user.
    """

    api_client_id: str
    api_client_secret: SecretStr
    url: str
    log_file: str = "bulk_restore.log"
    log_level: str = "INFO"
    _logger: Optional[logging.Logger] = None
    _device_loggers: Optional[dict] = None

    def __init__(self, **kwargs):
        # clear any keys from kwargs that are passed as None, which allows fallback to values in ENV
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        super().__init__(**kwargs)

    @validator("log_file")
    def validate_log_file(cls, value):  # noqa
        if "~" in value:
            value = str(value.replace("~", str(Path.home())))

        # will raise if not valid for writing but won't create file
        LazyFile(value, "a")

        return value

    @property
    def logger(self):
        if self._logger is None:
            handler = logging.FileHandler(
                filename=self.log_file,
            )
            handler.setFormatter(brt_format)
            self._logger = logging.getLogger("main")
            self._logger.setLevel(self.log_level)
            self._logger.addHandler(handler)
        return self._logger

    def get_device_loggers(self, device: Device, target: Path):
        if self._device_loggers is None:
            self._device_loggers = {}

        if device.deviceId not in self._device_loggers:
            device_logger = logging.getLogger(device.deviceId)
            device_filename = f"{device.username}_{device.name}_{device.deviceId}.log"
            device_handler = logging.FileHandler(filename=target / device_filename)
            device_handler.setFormatter(brt_format)
            device_logger.addHandler(device_handler)
            device_logger.handlers.extend(self.logger.handlers)
            device_logger.setLevel("INFO")

            error_logger = logging.getLogger(f"{device.deviceId}-error")
            error_filename = (
                f"{device.username}_{device.name}_{device.deviceId}_error.log"
            )
            error_handler = logging.FileHandler(filename=target / error_filename)
            error_handler.setFormatter(brt_format)
            error_logger.addHandler(error_handler)
            error_logger.handlers.extend(self.logger.handlers)
            error_logger.setLevel("INFO")

            self._device_loggers[device.deviceId] = (device_logger, error_logger)

        return self._device_loggers[device.deviceId]

    class Config:
        underscore_attrs_are_private = True
        env_prefix = "brt_"
        env_file = ".env"



from types import TracebackType
from typing import Optional, Type
from urllib.parse import urlparse

from rich.live import Live
from rich.progress import (
    Progress,
    SpinnerColumn,
    TextColumn,
    BarColumn,
    DownloadColumn,
    TimeElapsedColumn,
)
from rich.table import Table

from bulk_restore_tool.models import RestoreJob, RestorableFile


class ProgressBar:
    """Manages a group of progress bars for a restore job."""

    def __init__(self, job: RestoreJob):
        job.calculate_stats()
        self._job = job
        self._table = Table.grid()

        # progress bar for each device
        description_template = "Device guid={0} name={1}"
        justify = len(
            max(
                description_template.format(device.deviceId, device.name)
                for device in job.devices.values()
            )
        )
        for device in job.devices.values():
            device_table = Table.grid()
            device.progress = Progress(
                TextColumn("[progress.description]{task.description}"),
                BarColumn(bar_width=None),
                DownloadColumn(),
                expand=True,
            )
            description = description_template.format(
                device.deviceId, device.name
            ).ljust(justify, " ")
            device.task_id = device.progress.add_task(
                description=description,
                total=device.totalBytes,
                completed=device.completedBytes,
            )
            device_table.add_row(device.progress)
            device.archive_progress = Progress(
                TextColumn("  "),
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
            )
            device_table.add_row(device.archive_progress)
            device_table.add_row()
            self._table.add_row(device_table)

        # job total progress
        total_columns = [
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
        ]
        if len(job.devices) > 1:
            total_columns += [BarColumn(bar_width=None), DownloadColumn()]
        self._total_progress_bar = Progress(*total_columns, expand=len(job.devices) > 1)
        self._completedFiles = sum(d.completedFiles for d in job.devices.values())
        completedBytes = sum(d.completedBytes for d in job.devices.values())
        self._total_completed_task = self._total_progress_bar.add_task(
            description=f"Total files restored: [green]{self._completedFiles}[/green] Elapsed:",
            completed=completedBytes,
            total=job.totalByteCount,
        )
        self._table.add_row(self._total_progress_bar)

        # error counter
        self._error_count = 0
        self._total_errors_bar = Progress(
            TextColumn("[progress.description]{task.description}")
        )
        self._total_error_task = self._total_errors_bar.add_task(
            description=f"Total files failed: {self._error_count}"
        )
        self._table.add_row(self._total_errors_bar)

        # Live class manages refreshing progress bars together
        self._live = Live(self._table, refresh_per_second=10)
        self._file_tasks = {}  # type: ignore

    def start_file_progress(self, file: RestorableFile):
        device = self._job.devices[file.deviceId]
        node = urlparse(file.storageNodeWebUrl).netloc.split(":")[0]
        details = f"fileId={file.fileId} from archive={file.archiveId} @ {file.destinationName} - {node}, size={file.size}"
        if file.failure_count > 0:
            description = f"Retrying {details} - [red]retry {file.failure_count}[/red]"
        else:
            description = f"Restoring {details}"
        self._file_tasks[file.fileId] = device.archive_progress.add_task(
            description=description, start=False
        )

    def complete_file_progress(self, file: RestorableFile):
        device = self._job.devices[file.deviceId]
        device.archive_progress.remove_task(self._file_tasks[file.fileId])
        if file.completed:
            device.progress.update(device.task_id, advance=file.sourceLengthBytes)
            self._completedFiles += 1
            self._total_progress_bar.update(
                task_id=self._total_completed_task,
                description=f"Total files restored: [green]{self._completedFiles}[/green] Elapsed:",
                advance=file.sourceLengthBytes,
            )
        else:
            if file.failure_count > 2:
                self._error_count += 1
                self._total_errors_bar.update(
                    task_id=self._total_error_task,
                    description=f"Total files failed: [red]{self._error_count}",
                )

    def __enter__(self) -> "ProgressBar":
        self._live.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self._live.stop()

    def print(self, *args, **kwargs):
        self._live.console.print(*args, **kwargs)



import re
from pathlib import Path
from queue import Queue
from typing import List

import sqlite_utils

from bulk_restore_tool.models import Archive
from bulk_restore_tool.models import FileMetadata
from bulk_restore_tool.models import FileSelection
from bulk_restore_tool.models import RestorableFile

RESTORABLE_SQL = """
    SELECT
        a.deviceId,
        a.archiveId,
        a.destinationName,
        a.storageNodeWebUrl,
        f.fileId,
        f.path,
        f.backupSetId,
        f.versionTimemillis,
        max(f.sourceLastModifiedMillis) as sourceLastModifiedMillis,
        f.sourceLengthBytes,
        f.sourceSHA256Checksum
    FROM file f
    JOIN archive a on (a.archiveId = f.archiveId)
    WHERE f.fileType != 'DIRECTORY'
    AND f.selected
    AND NOT f.completed
    GROUP BY f.path
"""


def save_archive_metadata_sqlite(
    archive: Archive,
    files: List[dict],
    target: Path,
    selection: FileSelection,
) -> Path:
    file_path = target / archive.db_filename
    db = sqlite_utils.Database(file_path)
    archive_table = db["archive"].create(  # type: ignore
        columns={name: field.type_ for name, field in Archive.__fields__.items()},
        pk="archiveId",
        if_not_exists=True,
    )
    file_table = db["file"].create(  # type: ignore
        columns={name: field.type_ for name, field in FileMetadata.__fields__.items()},
        pk=("archiveId", "fileId"),
        if_not_exists=True,
    )
    archive_table.upsert(archive.dict(), pk="archiveId")
    include_patterns = [re.compile(pat) for pat in selection.includePatterns]
    exclude_patterns = [re.compile(pat) for pat in selection.excludePatterns]
    for file in files:
        model = FileMetadata(**file, archiveId=archive.archiveId)

        # apply include patterns
        if include_patterns:
            for regex in include_patterns:
                if regex.match(model.path):
                    model.selected = True
                    model.excludeReason = None
        else:
            model.selected = True
            model.excludeReason = None

        # skip other filter logic if file isn't included
        if not model.selected:
            model.excludeReason = "Didn't match any include patterns."
            file_table.upsert(model.dict(), pk=("archiveId", "fileId"))
            continue

        # apply exclude patterns
        if exclude_patterns:
            for regex in exclude_patterns:
                if regex.match(model.path):
                    model.selected = False
                    model.excludeReason = f"Matched exclude pattern: {regex.pattern}"

        # apply backup set filter
        if selection.backupSetIds:
            if model.backupSetId not in selection.backupSetIds:
                model.selected = False
                model.excludeReason = (
                    f"Not in selected backupSetIds: {selection.backupSetIds}"
                )

        # apply deleted file age filter
        if selection.includeDeleted and selection.deletedAgeMillis is not None:
            if model.deleted and model.deletionTimeMillis < selection.deletedAgeMillis:
                model.selected = False
                model.excludeReason = (
                    f"Delete prior to maxDeletedAge: {selection.deletedMaxAge}"
                )

        file_table.upsert(model.dict(), pk=("archiveId", "fileId"))

    db.create_view("restorable", RESTORABLE_SQL, replace=True)
    return file_path


def restore_queue(db: sqlite_utils.Database, archive_id: str) -> Queue[RestorableFile]:
    queue: Queue[RestorableFile] = Queue()
    for file in db.query(
        """
        SELECT *
        FROM restorable
        WHERE archiveId = :archiveId
        ORDER BY sourceLengthBytes ASC
        """,
        {"archiveId": archive_id},
    ):
        queue.put(RestorableFile(**file))
    return queue
